/* asm9830: a portable cross-assembler for the HP 9830A/B calculator */
/* Copyright (c) 2013-2014, Robert Ferguson (rob@bitscience.ca) */
/* All rights reserved; see LICENSE for details */

#include <stdio.h>  
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/errno.h>

#include "parser.tab.h"
#include "asm9830.h"
     
extern int nerror;

/* show the context of a warning/error in the source file */

static void
showcontext(YYLTYPE *locp)
{
	FILE *fp;
	int i, column = 0;
	char buf[BUFSIZ];

	/* open the input file where the error occurred, and seek to the correct location */

	if ((fp = fopen(locp->filename, "r")) == NULL)
		return;

	/* print the source line */

	fseek(fp, locp->lineoffset, SEEK_SET);
	fgets(buf, BUFSIZ, fp);
	fprintf(stderr, "\t%s\t", buf);

	/* align the error pointer under offending part */

	for (i = 0; i < locp->first_column; i++) {
		if (buf[i] == '\t')
			column += 8 - ((column + 8) % 8);
		else
			column++;
	}

	fprintf(stderr, "%*c^\n", column - 1, ' ');
	fclose(fp);
}

/* yyparse/codegen errors w/optional program context */

void 
yyerror(YYLTYPE *locp, Program *pgm, char const *msg, ...)
{
	va_list ap;
	va_start(ap, msg);

	nerror++;

	if (locp->first_line) {
		fprintf(stderr, "%s:%d.%d: error: ", (locp->filename == NULL) ? "stdin" : locp->filename, 
			locp->first_line, locp->first_column);
	}

	vfprintf(stderr, msg, ap);
	fprintf(stderr, "\n");

	if (verbose && locp->first_line && locp->filename) {
		showcontext(locp);
	}
}

/* warnings w/optional program context. Only show warnings when command line warnings flag set */

void
warning(YYLTYPE *locp, char const *msg, ...)
{
	va_list ap;
	va_start(ap, msg);

	if (!warnings)
		return;

	if (locp->first_line) {
		fprintf(stderr, "%s:%d.%d: warning: ", (locp->filename == NULL) ? "stdin" : locp->filename, 
			locp->first_line, locp->first_column);
	}

	vfprintf(stderr, msg, ap);
	fprintf(stderr, "\n");

	if (verbose && locp->first_line && locp->filename) {
		showcontext(locp);
	}
}

/* general error messages */

void
error(char *s, ...)
{
	va_list ap;

	nerror++;

	va_start(ap, s);
	vfprintf(stderr, s, ap);
	va_end(ap);

	if (errno) {
		fprintf(stderr, ": %s", strerror(errno));
	}

	printf("\n");
}
