/* asm9830: a portable cross-assembler for the HP 9830A/B calculator */
/* Copyright (c) 2013-2014, Robert Ferguson (rob@bitscience.ca) */
/* All rights reserved; see LICENSE for details */

#include <stdio.h>
#include <stdlib.h>

#include "parser.tab.h"
#include "asm9830.h"

static Expr *
createexpr(YYLTYPE loc)
{
	Expr *t;

	if ((t = (Expr *)malloc(sizeof(Expr))) == NULL)
		return(NULL);

	t->type = EXPR_VALUE;
	t->value = 0;
	t->left = NULL;
	t->right = NULL;
	t->next = NULL;
	t->loc = loc;

	return(t);
}

Expr *
numexpr(YYLTYPE loc, int value)
{
	Expr *t = createexpr(loc);

	if (t == NULL)
		return(NULL);

	t->type = EXPR_VALUE;
	t->value = value;
	return(t);
}

Expr *
symbolexpr(YYLTYPE loc, Symbol *s)
{
	Expr *t = createexpr(loc);

	if (t == NULL)
		return(NULL);

	t->type = EXPR_SYMBOL;
	t->s = s;

	return(t);
}

Expr *
opexpr(YYLTYPE loc, int op, Expr *l, Expr *r)
{
	Expr *t = createexpr(loc);

	if (t == NULL) 
		return(NULL);

	t->type = EXPR_OP;
	t->value = op;
	t->left = l;
	t->right = r;

	return(t);
}

/*
 * Preserve *perr across calls, so that errors accumulate over a series of recursions;
 * eloc gets the location of the error(s).
 *
 * If a page number is supplied, do symbol evaluations with page relative addressing
 * when applicable. 
 */

int
evalexpr(Expr *t, int page, int *perr, YYLTYPE *eloc) 
{
	int left, right;

	if (t == NULL) {
		*perr |= EXPR_NULL;
		return(0);
	}

	switch(t->type) {
	case (EXPR_VALUE):
		return(t->value);
		break;
	case (EXPR_SYMBOL):
		if (t->s->type == SYMBOL_FORWARD) {
			*perr |= EXPR_SYMBOL_FORWARD;
			*eloc = t->loc;
			return(0);
		}
		
		if (page && (PAGEINDEX(t->s->value) == page)) {
			return(PAGEOFFSET(t->s->value));
		} else {
			return(t->s->value);
		}
		break;
	case (EXPR_OP):
		/* unary ops don't have a left argument */
		if (t->left)
			left = evalexpr(t->left, page, perr, eloc);

		right = evalexpr(t->right, page, perr, eloc);
		
		switch(t->value) {
		case (OP_PLUS):
			return(left + right);
		case (OP_MINUS):
			return(left - right);
		case (OP_MUL):
			return(left * right);
		case (OP_DIV):
			return(left / right);
		case (OP_SHL):
			return(left << right);
		case (OP_SHR):
			return(left >> right);
		case (OP_AND):
			return(left & right);
		case (OP_OR):
			return(left | right);
		case (OP_XOR):
			return(left ^ right);
		case (OP_NEG):
			return(-right);
		case (OP_TWDL):
			return(~right);
		default:
			*perr |= EXPR_ERROR;
			*eloc = t->loc;
		}
		break;
	default:
		*perr |= EXPR_ERROR;
		*eloc = t->loc;
		break;
	}
	return(0);		
}

static char *exprerrormsg[] = {
	"unresolved reference",
	"malformed expression or arithmetic error",
	"null expression"
};

char *
errexpr(int err)
{
	if (err & EXPR_SYMBOL_FORWARD) 
		return(exprerrormsg[0]);

	if (err & EXPR_ERROR)
		return(exprerrormsg[1]);

	if (err & EXPR_NULL)
		return(exprerrormsg[2]);
	
	return(NULL);
}

