/* asm9830: a portable cross-assembler for the HP 9830A/B calculator */
/* Copyright (c) 2013-2014, Robert Ferguson (rob@bitscience.ca) */
/* All rights reserved; see LICENSE for details */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "parser.tab.h"
#include "asm9830.h"
#include "opbits.h"
#include "header.h"

/* evaluate/distribute the bits in the instruction data into the appropriate prototype */

static unsigned short
_makebits(Inst *i) 
{
	int out = 0, outbit;
	int value, valuebit;
	int err = 0, p;
	char c;
	char *proto = opbits[i->opcode];
	YYLTYPE eloc;

	value = evalexpr(i->argument, 0, &err, &eloc);

	/* handle pseudo ops */

	if (i->opcode == OPCODE_DATA) {
		if (err != 0) {
			yyerror(&eloc, NULL, "codegen: error evaluating argument: %s", errexpr(err));
			return(0);
		}
		if (value > ((1 << 16) - 1)) {
			yyerror(&eloc, NULL, "codegen: OPCODE_DATA argument out of range; value = 0%o", value);
			return(0);
		} else 
			return(value);
	}

	/* evaluate and validate expression argument */

	if (strpbrk(proto, "mbdn") && (err != 0)) {
		yyerror(&eloc, NULL, "codegen: error evaluating argument: %s", errexpr(err));
		return(0);
	}

	/* memory address must fit in 10 bits; must be on zero-page or current-page */

	if (strchr(proto, 'm')) {

		if (value > ((1 << 11) - 1)) {

			if ((PAGEINDEX(i->address) == PAGEINDEX(value))) {
				// value is in the current page; use only lower 10 bits
				value = value & ((1 << 11) - 1);
				i->modifier |= MOD_CURRENT;
			} else {
				yyerror(&(i->argument->loc), NULL, "codegen: address not on current or zero page; value = 0%o", value);
				return(0);
			}
		} else {
			// value is implicitly on zero page
			// clear the current page bit in an abundance of paranoia.
			i->modifier &= ~MOD_CURRENT;
		}
	}

	/* value of branch offset must be less than 5 bits, including sign bit */

	if (strchr(proto, 'b')) {
		value = value - i->address;
		if (abs(value) > ((1 << 5) - 1)) {
			yyerror(&(i->argument->loc), NULL, "codegen: branch offset too large: 0%o", value);
			return(0);
		}
	}

	/* i/o device channel must fit in 5 bits */

	if (strchr(proto, 'd')) {
		if (value > ((1 << 6) - 1)) {
			yyerror(&(i->argument->loc), NULL, "codegen: i/o device channel number too large: 0%o", value);
			return(0);
		}
	}

	/* shift count must fit in 4 bits */

	if (strchr(proto, 'n')) {

		/* "For right shift instructions, the shift count is n+1; for left-shift, 16-n" */

		if (i->opcode == OPCODE_SRL) {
			value = 16 - value;
		} else {
			value = value - 1;
		}

		if (value > ((1 << 5) - 1)) {
			yyerror(&(i->argument->loc), NULL, "codegen: shift count too large: 0%o", value);
			return(0);
		}
	}

	/* walk the prototype string, and set the bits in the result */

	valuebit = 0;
	outbit = 0;

	for (p = strlen(proto); p >= 0; p--) {
		c = proto[p];

		switch(c) {
		case('0'):
		case('1'):
			out |= (c - '0') << outbit++;
			break;

		case('.'):
			outbit++;
			break;

		case('r'):
			out |= (i->opreg) << outbit++;
			break;

		case('g'):
			out |= (i->argreg) << outbit++;
			break;

		case('i'):
			out |= ((i->modifier & MOD_INDIRECT) != 0) << outbit++;
			break;

		case('p'):
			out |= ((i->modifier & MOD_CURRENT) != 0) << outbit++; 
			break;

		case('s'):
			out |= ((i->modifier & MOD_SET) != 0) << outbit++;
			break;

		case('c'):
			out |= ((i->modifier & MOD_CLEAR) != 0) << outbit++;
			break;

		case('m'):
		case('b'):
		case('d'):
		case('n'):
			out |= ((value & (1 << valuebit++)) != 0) << outbit++;
			break;

		default:
			/* space, other unrecognized character */
			break;
		}
	}

	return(out);
}


static int
_emit(unsigned short *in, int words, FILE *fp, int isxns) 
{
	static int position = 0;
	char buf[6];
	int i, n;
	char *hex = "0123456789ABCDEF";

	if (!isxns) {
		// the HP 9830 is a little-endian machine. If byte order needs to 
		// be swapped, this would be the place to do it.

		return(fwrite(in, sizeof(unsigned short), words, fp)); 
	} else {
		for (i = 0; i < words; i++) {
			n = 0; 
			buf[n++] = hex[((in[i]>>12) & 0xf)];
			buf[n++] = hex[((in[i]>>8) & 0xf)];
			buf[n++] = hex[((in[i]>>4) & 0xf)];
			buf[n++] = hex[(in[i] & 0xf)];
			buf[n++] = '~';
			if ((++position % 10) == 0) {
				buf[n++] = '\n';
			}
			if (fwrite(buf, 1, n, fp) != n) {
				error("codegen: output file write failed");
				return(-1);
			}
		}
		return(words);
	}
}


int
codegen(Program *pgm, FILE *output, FILE *listing, int format)
{
	Inst *i;
	unsigned short o;
	FILE *fp = NULL;
	char *current = NULL;
	char line[BUFSIZ];
	int last;
	unsigned short sum = 0;
	HP9830Header header;

	/* header */

	if (format & FMT_HEADER) {

		bzero(&header, sizeof(header));

		// cursz is the size of the program, not including the checksum

		header.magic = HP9830_MAGIC;
		header.ftype = HP9830_FT_BINARY;	
		header.cursz = pgm->count;	

		// TODO: should we generate a header checksum?

		if (_emit(header.data, sizeof(header.data)/sizeof(unsigned short), output, format & FMT_XNS) != 
				(sizeof(header.data)/sizeof(unsigned short))) {
			error("codegen: output file write failed: header");
			return(-1);
		}
	}

	/* program */

	for (i = pgm->p; i != NULL; i = i->next) {

		o = _makebits(i);

		sum += o;

		if (listing) {

			if (current && strcmp(i->loc.filename, current)) {
				fclose(fp);
				fp = NULL;
				current = NULL;
				fprintf(fp, "\n");
			}

			if (!fp) {
				if ((fp = fopen(i->loc.filename, "r")) == NULL) {
					error("can't open source file %s for listing", i->loc.filename);
					return(-1);
				}
				fprintf(listing, "[Source file '%s':]\n", i->loc.filename);
				current = i->loc.filename;
				last = -1;
			}

			/*
			 * Print all source lines from the last one that generated code, then print the 
			 * new code and source line.
			 *
			 * If the current instruction doesn't advance the source offset, that means a single 
			 * line of source generated multiple instructions; in that case, just print the 
			 * generated code without a source line.
		 	 */

			if (last != i->loc.lineoffset) {
				while (ftell(fp) < i->loc.lineoffset) {
					fgets(line, BUFSIZ, fp);
					fprintf(listing, "            \t\t%s", line);
				}
				fgets(line, BUFSIZ, fp);
				fprintf(listing, "%04o: %06ho\t\t%s", i->address, o, line);
				last = i->loc.lineoffset;
			} else {
				fprintf(listing, "%04o: %06ho\n", i->address, o);
			}
		}

		if (_emit(&o, 1, output, format & FMT_XNS) != 1) {
			error("codegen: output file write failed: body");
			return(-1);	
		}
	}

	if ((format & FMT_HEADER) && (_emit(&sum, 1, output, format & FMT_XNS) != 1)) {
		error("codegen: output file write failed: checksum");
		return(-1);	
	}

	// terminate the XNS output

	if (format & FMT_XNS) {
		fprintf(output, "\n");
	}
		
	if (fp) {
		fclose(fp);
	}

	return(0);
}
