/* asm9830: a portable cross-assembler for the HP 9830A/B calculator */
/* Copyright (c) 2013-2014, Robert Ferguson (rob@bitscience.ca) */
/* All rights reserved; see LICENSE for details */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "hash.h"

/* 
 * A simplistic 2-choice hashing implementation, aka "the power of two choices".
 *
 * See https://en.wikipedia.org/wiki/2-choice_hashing and pretty much all of 
 * Michael Mitzenmacher's research (http://www.eecs.harvard.edu/~michaelm/).
 *
 * TLDR: massive improvement in expected size of the largest bucket over 
 * standard hashing -- O(log n) to O(log log n) 
 */

/* djb's hash function */

static unsigned int
_hash1(char *key)
{
	unsigned long hash = 5381;
	int c;

	/* hash(i) = hash(i-1) * 33 + c */

	while ((c = *key++))
		hash = ((hash << 5) + hash) + c; 

	return(hash);
}

/* sdbm/gawk/berkeley db/etc. hash function */

static unsigned int
_hash2(char *key)
{
        unsigned long hash = 0;
        int c;

	/* hash(i) = hash(i-1) * 65559 + c */

        while ((c = *key++))
            hash = c + (hash << 6) + (hash << 16) - hash;  

        return(hash);
}

HTable *
createht(int n)
{
	HTable *ht;
	_Bucket *a;
	int i;

	if ((ht = (HTable *)malloc(sizeof(HTable))) == NULL) {
		printf("malloc failed\n");
		return(NULL);
	}

	if ((a = (_Bucket *)malloc(sizeof(_Bucket) * n)) == NULL) {
		printf("malloc failed\n");
		return(NULL);
	}

	ht->array = a;
	ht->n = n;

	for (i = 0; i < n; i++) {
		ht->array[i].list = NULL;
		ht->array[i].count = 0;
	}

	return(ht);
}

void
destroyht(HTable *ht, int dflag)
{
	_Entry *t, *next;
	int i;

	for (i = 0; i < (ht->n); i++) {
		t = ht->array[i].list;
		while (t) {
			next = t->next;

			if (dflag) {
				free(t->key);
				free(t->data);
			}

			free(t);
			t = next;
		}
	}
			
	free(ht->array);
	free(ht);
}

void *
insertht(HTable *ht, char *key, void *ptr) 
{
	unsigned int hv1, hv2;
	_Entry *e, *t;

	if ((e = (_Entry *)malloc(sizeof(_Entry))) == NULL) {
		return(NULL);
	}
	e->data = ptr;
	e->key = key;
	e->next = NULL;

	hv1 = _hash1(key) % ht->n;
	hv2 = _hash2(key) % ht->n;

	if (ht->array[hv1].count >  ht->array[hv2].count) {
		hv1 = hv2;
	}

	t = ht->array[hv1].list;
	ht->array[hv1].list = e;
	e->next = t;
	ht->array[hv1].count++;

	return(ptr);
}

void * 
deleteht(HTable *ht, char *key)
{
	unsigned int hv[2], hash;
	_Entry *t, *prev;
	int i;
	void *data;
	
	hv[0] = _hash1(key) % ht->n;
	hv[1] = _hash2(key) % ht->n;

	for (i = 0; i < 2; i++) {
		hash = hv[i];

		t = ht->array[hash].list;
		prev = NULL;

		while ((t != NULL) && (strcmp(t->key, key) != 0)) {
			prev = t;
			t = t->next;
		}

		if (t != NULL) {
			if (prev == NULL) {
				ht->array[hash].list = t->next;
			} else {
				prev->next = t->next;
			}

			ht->array[hash].count--;
			data = t->data;
		 	free(t);
			return(data);
		}
	}
	return(NULL);
}

void *
findht(HTable *ht, char *key) 
{
	unsigned int hv[2], hash;
	int i;
	_Entry *t;

	hv[0] = _hash1(key) % ht->n;
	hv[1] = _hash2(key) % ht->n;

	for (i = 0; i < 2; i++) {
		hash = hv[i];
		t = ht->array[hash].list;
		while ((t != NULL) && (strcmp(t->key, key) != 0)) {
			t = t->next;
		}

		if (t != NULL) {
			return(t->data);
		}
	}
	return(NULL);
}

typedef struct _htenumcontext {
	int bucket;
	_Entry *e;
} HTEnumContext;

void *
enumht(HTable *ht, char **key, void **context)
{
	HTEnumContext *ec;

	if (*context == NULL) {
		if ((ec = malloc(sizeof(HTEnumContext))) == NULL)
			return(NULL);

		ec->bucket = 0;
		ec->e = NULL;

		*context = ec;
	} else {
		ec = *context;
	}

	while (ec->bucket < ht->n ) {
		if (ec->e == NULL) {
			ec->e = ht->array[ec->bucket].list;
		} else {
			ec->e = ec->e->next;
		}

		if (ec->e != NULL) {
			*key = ec->e->key;
			return(ec->e->data);
		}

		ec->bucket++;
	}

	free(*context);
	return(NULL);
}

int
countht(HTable *ht)
{	
	int i, sum = 0;

	for (i = 0; i < ht->n; i++)
		sum += ht->array[i].count;

	return(sum);
}

#ifdef DEBUG
static void
_dumpht(HTable *ht)
{
	int i;
	int high, low;

	high = 0;
	low = ht->array[0].count;

	for (i = 0; i < ht->n; i++) {
		printf("bucket %d: %d\n", i, ht->array[i].count);
		if (ht->array[i].count > high) {
			high = ht->array[i].count;
		} else {
			if (ht->array[i].count < low) {
				low = ht->array[i].count;
			}
		}
	}
	printf("low = %d, high = %d\n", low, high);
}
#endif
