/* asm9830: a portable cross-assembler for the HP 9830A/B calculator */
/* Copyright (c) 2013-2014, Robert Ferguson (rob@bitscience.ca) */
/* All rights reserved; see LICENSE for details */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "parser.tab.h"
#include "asm9830.h"


/* 
 * maintain a queue of pending instructions during intermediate stages of parsing: 
 * once parsing of source line completed, instructions are added to instruction stream 
 * by commitinst() or unwound by abortinst().
 */

Inst *
queueinst(Program *pgm, YYLTYPE loc, int opcode, int opreg, int argreg, Expr *arg, int mod)
{
	Inst *t, *i;
	
	if ((t = (Inst *)malloc(sizeof(Inst))) == NULL) {
		error("can't allocate instruction record\n");
		exit(-1);
	}
	
	t->loc = loc;
	t->opcode = opcode;
	t->address = -1;
	t->argreg = argreg;
	t->opreg = opreg;
	t->argument = arg;
	t->modifier = mod;
	t->next = NULL;

	if (pgm->queue == NULL) {
		pgm->queue = t;
	} else {
		for (i = pgm->queue; i->next != NULL; i = i->next) 
			;

		i->next = t;
	}

	return(t);
}

void
commitinst(Program *pgm)
{
	Inst *i;

	if (pgm->dot == -1) {
		pgm->dot = DEFAULT_ORIGIN;
	}

	if (pgm->last == NULL) {
		pgm->p = pgm->queue;
	} else {
		pgm->last->next = pgm->queue;
	}

	/* special case to commit ORG "instruction"; new dot value stored in mod field */

	for (i = pgm->queue; i != NULL; i = i->next) {
		if (i->opcode == OPCODE_ORG) {
			pgm->dot = i->modifier;
		} else {
			i->address = pgm->dot++;
			pgm->last = i;
			pgm->count++;
		}
	}

	pgm->queue = NULL;
}


void
abortinst(Program *pgm)
{
	Inst *i = pgm->queue, 
	     *n;

	while (i) {
		n = i->next;
		free(i);
		i = n;
	}

	pgm->queue = NULL;
}

/* check a set of instruction modifiers against the acceptable set */

#define MSGSZ 32

struct _namemap {
	int value;
	char *name;
}; 

static struct _namemap modnames[] = {
	{MOD_INDIRECT, "Indirect "},
	{MOD_CLEAR, "Clear "},
	{MOD_SET, "Set "},
};

int
checkmod(YYLTYPE loc, int arg, int mod) 
{
	int i;
	char modlist[MSGSZ] = "";
	char arglist[MSGSZ] = "";

	/* are there bits in arg not in mod? */

	if (((arg ^ mod) & arg) != 0) {

		for (i = 0; i < sizeof(modnames)/sizeof(struct _namemap); i++) {

			if (arg & modnames[i].value) {
				strlcat(arglist, modnames[i].name, MSGSZ);
			}
			if (mod & modnames[i].value) {
				strlcat(modlist, modnames[i].name, MSGSZ);
			}
		}
		yyerror(&(loc), NULL, "incorrect instruction Modifier; expected '%s', found  '%s'", modlist, arglist); 
		return(0);
	}
	return(1);
} 

void
dumpinst(Inst *i)
{
	unsigned short result; 
	int error = 0;
	YYLTYPE eloc;

	printf("\t(%s)(%d.%d): address = 0%o; opcode = 0%o, opreg = 0%o\n", 
		i->loc.filename, i->loc.first_line, i->loc.first_column, 
		i->address, i->opcode, i->opreg);

	printf("\targreg = 0%o, argument = %p, modifier = 0%o\n", 
		i->argreg, i->argument, i->modifier);

	result = evalexpr(i->argument, 0, &error, &eloc);
	if (error == 0) {
		printf("\tevalexpr(argument) = %hd (0%06ho), ", result, result);
	} else {
		printf("\targument evaluates with error %d, ", error);
	}
	printf("\n");
}

void 
dumppgm(Program *pgm)
{
	Inst *i;

	for (i = pgm->p; i != NULL; i = i->next) {
		dumpinst(i);
	}
}
