/* asm9830: a portable cross-assembler for the HP 9830A/B calculator */
/* Copyright (c) 2013-2014, Robert Ferguson (rob@bitscience.ca) */
/* All rights reserved; see LICENSE for details */

#include <stdio.h>  
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <fcntl.h>
#include <unistd.h>
#include <sys/errno.h>

#include "parser.tab.h"
#include "asm9830.h"
     
extern int yydebug;
extern FILE *yyin;     
extern char *yyfilename;
extern unsigned long yyoffset;

int warnings = 0;
int verbose = 0;
int nerror = 0;

void
usage(char *name)
{
	fprintf(stderr, "usage: %s [-v] [-h] [-x] [-d] [-a listfile] [-o outfile] ([-i] | [infile ...])\n", name);
}

int
main (int argc, char *argv[])
{
	char *pgmname = argv[0];
	char *output = NULL;
	char *listing = NULL;
	FILE *fpout = NULL;
	FILE *fplist = NULL;
	int usestdin = 0;
	int format = FMT_HEADER;
	int i;
	int c, r;
	STable *stlocal, *stpublic;
	Program program;

	while ((c = getopt(argc, argv, "hxvda:io:")) != -1) {
		switch (c) {
		case('v'):
			verbose = 1;
			break;
		case('x'):
			format |= FMT_XNS;
			break;
		case('a'):
			listing = optarg;
			break;
		case('i'):
			usestdin = 1;
			break;
		case('o'):
			output = optarg;
			break;
		case('h'):
			format &= ~FMT_HEADER;
			break;
		case('d'):
			yydebug = 1;
			break;
		case('?'):
		default:
			usage(pgmname);
			return(-1);
		}
	}

	argc -= optind;
	argv += optind;

	if (usestdin && (argc > 0)) {
		error("non-option arguments follow -i\n");
		usage(pgmname);
		return(-1);
	}

	if (!usestdin && (argc == 0)) {
		error("missing input filename(s), and -i not specified");
		usage(pgmname);
		return(-1);
	}

	/* setup symbol tables and initialize program structure */

	stpublic = createst(NULL);
	stlocal = createst(stpublic);

	program.p = NULL;
	program.last = NULL;
	program.queue = NULL;
	program.st = stlocal;
	program.dot = -1;
	program.count = 0;

	if (listing) {
		if ((fplist = fopen(listing, "w")) == NULL) {
			error("Can't open listing file %s", listing);
			return(-1);
		}
	}

	if (argc > 0) {
		for (i = 0; i < argc; i++) {
			if (((yyin = fopen(argv[i], "r")) == 0)) {
				error("Can't open input file %s\n", argv[i]);
				return(-1);
			} 
			
			yyfilename = argv[i];
			yyoffset = 0;
			if ((r = yyparse(&program)) != 0) {
				error("yyparse returned %d", r);
				return(-1);
			}

			fclose(yyin);

			if (fplist) {
				fprintf(fplist, "[Local symbol table -- %s:]\n", argv[i]);
				listst(program.st, fplist);
				fprintf(fplist, "\n");
			}

			/* remove local symbol table, but preserve global/public one across files */
			/* FIXME: this leaks symbols */

			destroyst(program.st);
			program.st = createst(stpublic);
		}
	} else {
		yyfilename = NULL;
		yyoffset = 0;
		if ((r = yyparse(&program)) != 0) {
			error("yyparse returned %d", r);
		}
	}

	if (fplist) {
		fprintf(fplist, "[Global symbol table:]\n");
		listst(stpublic, fplist);
		fprintf(fplist, "\n");
	}

	if (output == NULL) {
		if (format & FMT_XNS) {
			output = DEFAULT_XNSOUT;
		} else { 
			output = DEFAULT_BINOUT;
		}
	}

	if ((fpout = fopen(output, "w")) == NULL) {
		error("Can't open output file %s", output);
		return(-1);
	}

	r = codegen(&program, fpout, fplist, format);

	fclose(fpout);
	if (fplist)
		fclose(fplist);

	if (r) {
		remove(output);
		return(-1);
	}

	if (nerror) {
		error("encountered %d errors during assembly", nerror);
		return(-1);
	} 

	return(0);
}
