/* asm9830: a portable cross-assembler for the HP 9830A/B calculator */
/* Copyright (c) 2013-2014, Robert Ferguson (rob@bitscience.ca) */
/* All rights reserved; see LICENSE for details */

/* 
 * Not entirely happy about C99 initializers, but it's the best way 
 * to keep the OPCODE_ defines and the array indexes in sync. 
 */

/* 
 * these instruction prototypes follow those in Brent's instruction set analysis; 
 * they are written from MSB to LSB, spaces are ignored, 0 and 1 are literal,
 * and the meaning of the other letters is set out below.
 * 
 * Key:
 *
 * r : A=0 / B=1 register                   m : bit of 10-bit memory address
 * g : A=0 / B=1 register                   b : bit of 5-bit branch offset
 * i : Direct=0 / Indirect=1 addressing     d : bit of 5-bit I/O device channel
 * p : Zero=0 / Current=1 page              n : bit of 4-bit shift count 
 * s : Hold=0 / Set=1 applicable bit        . : not used (assumed 0)
 * c : Hold=0 / Clear=1 applicable bit  
 *
 * See http://www.cs.ubc.ca/~hilpert/e/HP9830/machine.html for more details.
 */

char *opbits[] = {

/* Memory Reference Group: */

[OPCODE_ADR] = 	"i 000 rpm mmm mmm mmm",
[OPCODE_CPR] =	"i 001 rpm mmm mmm mmm",
[OPCODE_LDR] =	"i 010 rpm mmm mmm mmm",
[OPCODE_STR] =	"i 011 rpm mmm mmm mmm",
[OPCODE_IOR] =	"i 100 0pm mmm mmm mmm",
[OPCODE_ISZ] =	"i 100 1pm mmm mmm mmm",
[OPCODE_AND] =	"i 101 0pm mmm mmm mmm",
[OPCODE_DSZ] =	"i 101 1pm mmm mmm mmm",
[OPCODE_JSM] =	"i 110 0pm mmm mmm mmm",
[OPCODE_JMP] =	"i 110 1pm mmm mmm mmm",

/* D/I Reference Group: */

[OPCODE_ADRg] =	"0 111 g.. i00 0r1 111",
[OPCODE_CPRg] =	"0 111 g.. i00 1r1 111",
[OPCODE_LDRg] =	"0 111 g.. i01 0r1 111",
[OPCODE_STRg] =	"0 111 g.. 101 1r1 111",
[OPCODE_IORg] =	"0 111 g.. i10 001 111",
[OPCODE_ISZg] =	"0 111 g.. 110 011 111",
[OPCODE_ANDg] =	"0 111 g.. i10 101 111",
[OPCODE_DSZg] =	"0 111 g.. 110 111 111",
[OPCODE_JSMg] =	"0 111 g.. 111 001 111",
[OPCODE_JMPg] =	"0 111 g.. i11 011 111",
[OPCODE_EXR] =	"0 111 r.. ... 001 110",
[OPCODE_DMA] =	"0 111 0.. ... 011 110",
[OPCODE_CMR] =	"0 111 r.. ... 101 110",
[OPCODE_TCR] =	"0 111 r.. ... 111 110",

/* Shift-Rotate Group: */

[OPCODE_ARR] =	"0 111 r.. nnn n.0 000",
[OPCODE_SRR] =	"0 111 r.. nnn n.0 010",
[OPCODE_SRL] =	"0 111 r.. nnn n.0 100",
[OPCODE_RRR] =	"0 111 r.. nnn n.0 110",

/* Alter-Skip Group: */

[OPCODE_SZR] =	"0 111 r0b bbb b01 000",
[OPCODE_RZR] =	"0 111 r1b bbb b01 000",
[OPCODE_SIR] =	"0 111 r0b bbb b11 000",
[OPCODE_RIR] =	"0 111 r1b bbb b11 000",
[OPCODE_SLR] =	"0 111 rsb bbb bc1 001",
[OPCODE_SRM] =	"0 111 rsb bbb bc1 010",
[OPCODE_SRP] =	"0 111 rsb bbb bc1 011",
[OPCODE_SES] =	"0 111 0sb bbb bc1 100",
[OPCODE_SEC] =	"0 111 0sb bbb bc1 101",

/* Input/Output Group: */

[OPCODE_STF] =	"1 111 .10 111 1dd ddd",
[OPCODE_CLF] =	"1 111 .11 111 1dd ddd",
[OPCODE_SFC] =	"1 111 .1c 111 0dd ddd",
[OPCODE_SFS] =	"1 111 .1c 101 0dd ddd",
[OPCODE_CLC] =	"1 111 .1c 101 1dd ddd",
[OPCODE_STC] =	"1 111 .1c 110 0dd ddd",
[OPCODE_OTR] =	"1 111 r1c 001 1dd ddd",
[OPCODE_LIR] =	"1 111 r1c 010 1dd ddd",
[OPCODE_MIR] =	"1 111 r1c 000 1dd ddd",

/* MAC Group: */

[OPCODE_RET] =	"1 111 000 100 000 010",
[OPCODE_MOV] =	"1 111 000 000 000 010",
[OPCODE_CLR] =	"1 111 000 000 000 000",
[OPCODE_XFR] =	"1 111 000 000 000 100",
[OPCODE_MRX] =	"1 111 100 100 011 000",
[OPCODE_MRY] =	"1 111 100 100 111 000",
[OPCODE_MLS] =	"1 111 001 100 000 000",
[OPCODE_DRS] =	"1 111 000 100 001 000",
[OPCODE_DLS] =	"1 111 101 100 000 000",
[OPCODE_FXA] =	"1 111 000 101 110 000",
[OPCODE_FMP] =	"1 111 001 100 110 000",
[OPCODE_FDV] =	"1 111 000 100 010 000",
[OPCODE_CMX] =	"1 111 100 100 000 000",
[OPCODE_CMY] =	"1 111 000 100 000 000",
[OPCODE_MDI] =	"1 111 000 101 100 000",
[OPCODE_NRM] =	"1 111 001 100 101 000" };
