	/* asm9830: a portable cross-assembler for the HP 9830A/B calculator */
	/* Copyright (c) 2013-2014, Robert Ferguson (rob@bitscience.ca) */
	/* All rights reserved; see LICENSE for details */

%require "3.0"
%code requires {

	#include "defs.h"

	/* add file name and seek offset to standard location, for better error reporting */

	typedef struct YYLTYPE {
		int first_line;
		int first_column;
		int last_line;
		int last_column;
		char *filename;
		unsigned long lineoffset;
	} YYLTYPE;

	#define YYLTYPE_IS_DECLARED	1

	#define YYLLOC_DEFAULT(Current, Rhs, N)				\
		do							\
			if (N)						\
			{						\
				(Current).first_line	= YYRHSLOC (Rhs, 1).first_line;	\
				(Current).first_column	= YYRHSLOC (Rhs, 1).first_column;\
				(Current).last_line	= YYRHSLOC (Rhs, N).last_line;	\
				(Current).last_column	= YYRHSLOC (Rhs, N).last_column; \
				(Current).lineoffset	= YYRHSLOC (Rhs, N).lineoffset; \
				(Current).filename	= YYRHSLOC (Rhs, 1).filename; \
			}						\
			else						\
			{						\
				(Current).first_line = (Current).last_line = 		\
				   YYRHSLOC (Rhs, 0).last_line;		\
				(Current).first_column = (Current).last_column =	\
				   YYRHSLOC (Rhs, 0).last_column;			\
				(Current).lineoffset = yylineoffset;	\
				(Current).filename = NULL;		\
			}						\
		while (0)
}
%define api.pure full
%define parse.lac full
%define parse.error verbose
%define parse.trace
%parse-param {Program *pgm} 
%locations

%union {
  int num;
  int opcode;
  int opreg;
  int argreg;
  int modifier;
  Inst *iptr;
  Expr *eptr;
  char *string;
}

%code {
	#include <stdio.h>  
	#include <math.h>   
	#include <stdlib.h>
	#include <string.h>
	#include "asm9830.h"

	int yylex(YYSTYPE * lvalp, YYLTYPE *llocp);

	/* extra location information for error reporting; used in the style of other yy... variables */

	char *yyfilename;		/* input filename */
	unsigned long yyoffset;		/* file seek offset of current location */
	unsigned long yylineoffset;	/* file seek offset of beginning of current line */
}


%token <opreg> T_ADR 
%token <opreg> T_CPR 
%token <opreg> T_LDR 
%token <opreg> T_STR
%token T_IOR
%token T_ISZ
%token T_AND
%token T_DSZ
%token T_JSM
%token T_JMP
%token <opreg> T_EXR
%token T_DMA
%token <opreg> T_CMR
%token <opreg> T_TCR
%token <opreg> T_ARR
%token <opreg> T_SRR
%token <opreg> T_SRL
%token <opreg> T_RRR
%token <opreg> T_SZR
%token <opreg> T_RZR
%token <opreg> T_SIR
%token <opreg> T_RIR
%token <opreg> T_SLR
%token <opreg> T_SRM
%token <opreg> T_SRP
%token T_SES
%token T_SEC
%token T_STF
%token T_CLF
%token T_SFC
%token T_SFS
%token T_CLC
%token T_STC
%token <opreg> T_OTR
%token <opreg> T_LIR
%token <opreg> T_MIR
%token T_RET
%token T_MOV
%token T_CLR
%token T_XFR
%token T_MRX
%token T_MRY
%token T_MLS
%token T_DRS
%token T_DLS
%token T_FXA
%token T_FMP
%token T_FDV
%token T_CMX
%token T_CMY
%token T_MDI
%token T_NRM

%token T_ORG
%token T_BSS
%token T_ASC
%token T_EQU
%token T_DW
%token T_DB
%token T_NOP

%token DOT
%token PUBLIC

%token <num>  NUM         
%token <string> LABEL 	
%token SHL SHR				/* shift left/right operators */

%token <modifier> MODIFIER;
%token <string> STRING;
%token <argreg> REG;

%type <iptr> line;
%type <iptr> instruction;
%type <eptr> expression;
%type <eptr> expression-list;

%left '+' '-'
%left '*' '/' 
%left '&' '|' '^' SHL SHR
%precedence NEG /* negation--unary minus */

%% 	

input:
	%empty
	| input line 
	| error line
	{ 
		yyerrok; 
	}
	;
     
line:
	LABEL instruction 
	{ 
		$$ = $2; 
		
		Symbol *s;

		if (((s = findsymbol(pgm->st, $1, SYMBOL_FOLLOW)) == NULL) || (s->type == SYMBOL_FORWARD)) {
			if (s == NULL) {
				s = addsymbol(pgm->st, $1, NULL, &@1);
			}
			if (pgm->dot == -1) {
				pgm->dot = DEFAULT_ORIGIN;
			}

			s->value = pgm->dot;
			s->type = SYMBOL_DEFINED;
			commitinst(pgm);
		} else {
			yyerror(&@1, NULL, "symbol %s previously defined; first use (%s)%d.%d", $1,
				s->loc.filename, s->loc.first_line, s->loc.first_column);
			abortinst(pgm);
			YYERROR;
		}
	}
	| PUBLIC LABEL instruction 
	{ 
		$$ = $3; 
		
		Symbol *s;

		/* 
		 * A 'public' symbol lives in the global symbol table; 
		 * if it's been previously defined (either locally or globally), 
		 * that's an error. 
		 *
		 * On the other hand, if it's undefined but has been used as a 
		 * forward reference, we need to move the symbol object from the 
		 * local to the global symbol table. 
		 */

		if ((s = findsymbol(pgm->st, $2, SYMBOL_NOFOLLOW)) != NULL) {
			if (s->type == SYMBOL_FORWARD) {
				/* delete the reference in the local symbol table, reinsert in public one */
				s = delsymbol(pgm->st, $2);
				addsymbol(pgm->st->parent, $2, s, NULL);
			} else {
				yyerror(&@2, NULL, "symbol %s declared 'public', but previously defined local; first use (%s)%d.%d",
					$2, s->loc.filename, s->loc.first_line, s->loc.first_column);
				abortinst(pgm);
				YYERROR;
			}
		} else {
			if (((s = findsymbol(pgm->st->parent, $2, SYMBOL_NOFOLLOW)) == NULL) || (s->type == SYMBOL_FORWARD)) {
				if (s == NULL) {
					s = addsymbol(pgm->st->parent, $2, NULL, &@2);
				}
			} else {
				yyerror(&@2, NULL, "symbol %s previously defined; first use (%s)%d.%d", 
					$2, s->loc.filename, s->loc.first_line, s->loc.first_column);
				abortinst(pgm);
				YYERROR;
			}
		}

		if (pgm->dot == -1) {
			pgm->dot = DEFAULT_ORIGIN;
		}

		s->value = pgm->dot;
		s->type = SYMBOL_DEFINED;
		commitinst(pgm);
	}
	| LABEL T_EQU expression '\n'
	{
		/* explicitly set label to be equal to the value of expression */
		/* expression must be resolvable; no forward references */

		Symbol *s;
		int value;
		int err = 0;
		YYLTYPE eloc;

		value = evalexpr($3, 0, &err, &eloc);

		if (err & EXPR_SYMBOL_FORWARD) {
			yyerror(&eloc, NULL, "EQU expression cannot be evaluated; forward reference");
			YYERROR;
		}

		if (err) {
			yyerror(&eloc, NULL, "EQU expression cannot be evaluated");
			YYERROR;
		}

		if ((s = findsymbol(pgm->st, $1, SYMBOL_FOLLOW)) != NULL) {
			if (s->type != SYMBOL_FORWARD) {
				yyerror(&@1, NULL, "symbol %s previously defined; first use (%s)%d.%d", $1,
					s->loc.filename, s->loc.first_line, s->loc.first_column);
				YYERROR;
			}
		} else {
			s = addsymbol(pgm->st, $1, NULL, &@1);
		}

		s->value = value;
		s->type = SYMBOL_DEFINED;

		$$ = NULL;
	}
	| PUBLIC LABEL T_EQU expression '\n'
	{
		/* as above, except handle public/private symbol table */

		Symbol *s;
		int value;
		int err = 0;
		YYLTYPE eloc;

		value = evalexpr($4, 0, &err, &eloc);
		
		if (err & EXPR_SYMBOL_FORWARD) {
			yyerror(&eloc, NULL, "EQU expression cannot be evaluated; forward reference");
			YYERROR;
		}

		if (err) {
			yyerror(&eloc, NULL, "EQU expression cannot be evaluated");
			YYERROR;
		}

		if ((s = findsymbol(pgm->st, $2, SYMBOL_NOFOLLOW)) != NULL) {
			if (s->type == SYMBOL_FORWARD) {
				/* delete the reference in the local symbol table, reinsert in public one */
				s = delsymbol(pgm->st, $2);
				addsymbol(pgm->st->parent, $2, s, NULL);
			} else {
				yyerror(&@2, NULL, "symbol %s declared 'public', but previously defined local; first use (%s)%d.%d",
					$2, s->loc.filename, s->loc.first_line, s->loc.first_column);
				YYERROR;
			}
		} else {
			if (((s = findsymbol(pgm->st->parent, $2, SYMBOL_NOFOLLOW)) == NULL) || (s->type == SYMBOL_FORWARD)) {
				if (s == NULL) {
					s = addsymbol(pgm->st->parent, $2, NULL, &@2);
				}
			} else {
				yyerror(&@2, NULL, "symbol %s previously defined; first use (%s)%d.%d", 
					$2, s->loc.filename, s->loc.first_line, s->loc.first_column);
				YYERROR;
			}
			s->value = value;
			s->type = SYMBOL_DEFINED;
		}
		
		$$ = NULL;
	}
	| instruction 
	{ 
		$$ = $1; 
		commitinst(pgm);
	}
	;


instruction:
	/* MAC instruction group */

	  T_RET '\n'	{ $$ = queueinst(pgm, @1, OPCODE_RET, 0, 0, 0, 0);  }
	| T_MOV '\n'	{ $$ = queueinst(pgm, @1, OPCODE_MOV, 0, 0, 0, 0);  }
	| T_CLR '\n'	{ $$ = queueinst(pgm, @1, OPCODE_CLR, 0, 0, 0, 0);  }
	| T_XFR '\n'	{ $$ = queueinst(pgm, @1, OPCODE_XFR, 0, 0, 0, 0);  }
	| T_MRX '\n'	{ $$ = queueinst(pgm, @1, OPCODE_MRX, 0, 0, 0, 0);  }
	| T_MRY '\n'	{ $$ = queueinst(pgm, @1, OPCODE_MRY, 0, 0, 0, 0);  }
	| T_MLS '\n'	{ $$ = queueinst(pgm, @1, OPCODE_MLS, 0, 0, 0, 0);  }
	| T_DRS '\n'	{ $$ = queueinst(pgm, @1, OPCODE_DRS, 0, 0, 0, 0);  }
	| T_DLS '\n'	{ $$ = queueinst(pgm, @1, OPCODE_DLS, 0, 0, 0, 0);  }
	| T_FXA '\n'	{ $$ = queueinst(pgm, @1, OPCODE_FXA, 0, 0, 0, 0);  }
	| T_FMP '\n'	{ $$ = queueinst(pgm, @1, OPCODE_FMP, 0, 0, 0, 0);  }
	| T_FDV '\n'	{ $$ = queueinst(pgm, @1, OPCODE_FDV, 0, 0, 0, 0);  }
	| T_CMX '\n'	{ $$ = queueinst(pgm, @1, OPCODE_CMX, 0, 0, 0, 0);  }
	| T_CMY '\n'	{ $$ = queueinst(pgm, @1, OPCODE_CMY, 0, 0, 0, 0);  }
	| T_MDI '\n'	{ $$ = queueinst(pgm, @1, OPCODE_MDI, 0, 0, 0, 0);  }
	| T_NRM '\n'	{ $$ = queueinst(pgm, @1, OPCODE_NRM, 0, 0, 0, 0);  }

	/* Input/output instruction group; modifier can be clear */

	| T_STF expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_STF, 0, 0, $2, 0);  }
	| T_CLF expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_CLF, 0, 0, $2, 0);  }
	| T_SFC expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_SFC, 0, 0, $2, 0);  }
	| T_SFC expression MODIFIER '\n'	
	{ 
		if (checkmod(@1, $3, MOD_CLEAR)) 
			$$ = queueinst(pgm, @1, OPCODE_SFC, 0, 0, $2, $3); 
		else 
			YYERROR; 
	}
	| T_SFS expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_SFS, 0, 0, $2, 0);  }
	| T_SFS expression MODIFIER '\n'	
	{ 
		if (checkmod(@1, $3, MOD_CLEAR)) 
			$$ = queueinst(pgm, @1, OPCODE_SFS, 0, 0, $2, $3); 
		else 
			YYERROR; 
	}
	| T_CLC expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_CLC, 0, 0, $2, 0);  }
	| T_CLC expression MODIFIER '\n'	
	{ 
		if (checkmod(@1, $3, MOD_CLEAR)) 
			$$ = queueinst(pgm, @1, OPCODE_CLC, 0, 0, $2, $3); 
		else 
			YYERROR; 
	}
	| T_STC expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_STC, 0, 0, $2, 0);  }
	| T_STC expression MODIFIER '\n'	
	{ 
		if (checkmod(@1, $3, MOD_CLEAR)) 
			$$ = queueinst(pgm, @1, OPCODE_STC, 0, 0, $2, $3); 
		else 
			YYERROR; 
	}
	| T_OTR expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_OTR, $1, 0, $2, 0);  }
	| T_OTR expression MODIFIER '\n'	
	{ 
		if (checkmod(@1, $3, MOD_CLEAR)) 
			$$ = queueinst(pgm, @1, OPCODE_OTR, $1, 0, $2, $3); 
		else 
			YYERROR; 
	}
	| T_LIR expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_LIR, $1, 0, $2, 0);  }
	| T_LIR expression MODIFIER '\n'	
	{ 
		if (checkmod(@1, $3, MOD_CLEAR))
			$$ = queueinst(pgm, @1, OPCODE_LIR, $1, 0, $2, $3); 
		else 
			YYERROR; 
	}
	| T_MIR expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_MIR, $1, 0, $2, 0);  }
	| T_MIR expression MODIFIER '\n'	
	{ 
		if (checkmod(@1, $3, MOD_CLEAR)) 
			$$ = queueinst(pgm, @1, OPCODE_MIR, $1, 0, $2, $3); 
		else 
			YYERROR; 
	}
	
	/* Alter-skip instruction group; modifier can be clear and/or set */

	| T_SZR expression '\n' 		{ $$ = queueinst(pgm, @1, OPCODE_SZR, $1, 0, $2, 0);  }
	| T_RZR expression '\n' 		{ $$ = queueinst(pgm, @1, OPCODE_RZR, $1, 0, $2, 0);  }
	| T_SIR expression '\n' 		{ $$ = queueinst(pgm, @1, OPCODE_SIR, $1, 0, $2, 0);  }
	| T_RIR expression '\n' 		{ $$ = queueinst(pgm, @1, OPCODE_RIR, $1, 0, $2, 0);  }
	| T_SLR expression '\n' 		{ $$ = queueinst(pgm, @1, OPCODE_SLR, $1, 0, $2, 0);  }
	| T_SLR expression MODIFIER '\n' 	
	{ 
		if (checkmod(@1, $3, MOD_CLEAR|MOD_SET)) 
			$$ = queueinst(pgm, @1, OPCODE_SLR, $1, 0, $2, $3); 
		else 
			YYERROR; 
	}
	| T_SRM expression '\n' 		{ $$ = queueinst(pgm, @1, OPCODE_SRM, $1, 0, $2, 0);  }
	| T_SRM expression MODIFIER '\n' 	
	{ 
		if (checkmod(@1, $3, MOD_CLEAR|MOD_SET)) 
			$$ = queueinst(pgm, @1, OPCODE_SRM, $1, 0, $2, $3); 
		else 
			YYERROR; 
	}
	| T_SRP expression '\n' 		{ $$ = queueinst(pgm, @1, OPCODE_SRP, $1, 0, $2, 0);  }
	| T_SRP expression MODIFIER '\n' 	
	{ 
		if (checkmod(@1, $3, MOD_CLEAR|MOD_SET)) 
			$$ = queueinst(pgm, @1, OPCODE_SRP, $1, 0, $2, $3); 
		else 
			YYERROR; 
	}
	| T_SES expression  '\n' 		{ $$ = queueinst(pgm, @1, OPCODE_SES, 0, 0, $2, 0);   }
	| T_SES expression MODIFIER '\n' 	
	{ 
		if (checkmod(@1, $3, MOD_CLEAR|MOD_SET)) 
			$$ = queueinst(pgm, @1, OPCODE_SES, 0, 0, $2, $3);  
		else 
			YYERROR; 
	}
	| T_SEC expression '\n' 		{ $$ = queueinst(pgm, @1, OPCODE_SEC, 0, 0, $2, 0);   }
	| T_SEC expression MODIFIER '\n' 	
	{ 
		if (checkmod(@1, $3, MOD_CLEAR|MOD_SET)) 
			$$ = queueinst(pgm, @1, OPCODE_SEC, 0, 0, $2, $3);  
		else 
			YYERROR; 
	}

	/* Shift/rotate instruction group */

	| T_ARR expression '\n' 	{ $$ = queueinst(pgm, @1, OPCODE_ARR, $1, 0, $2, 0);  }
	| T_SRR expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_SRR, $1, 0, $2, 0);  }
	| T_SRL expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_SRL, $1, 0, $2, 0);  }
	| T_RRR expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_RRR, $1, 0, $2, 0);  }

	/* D/I Reference instruction group; modifier can be indirect */

	| T_ADR REG '\n' 		{ $$ = queueinst(pgm, @1, OPCODE_ADRg, $1, $2, 0, 0); }
	| T_ADR REG MODIFIER '\n'	
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT)) 
			$$ = queueinst(pgm, @1, OPCODE_ADRg, $1, $2, 0, $3); 
		else 
			YYERROR; 
	}
	| T_CPR REG '\n'		{ $$ = queueinst(pgm, @1, OPCODE_CPRg, $1, $2, 0, 0); }
	| T_CPR REG MODIFIER  	
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT)) 
			$$ = queueinst(pgm, @1, OPCODE_CPRg, $1, $2, 0, $3); 
		else 
			YYERROR; 
	}
	| T_LDR REG '\n'		{ $$ = queueinst(pgm, @1, OPCODE_LDRg, $1, $2, 0, 0); }
	| T_LDR REG MODIFIER 	
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT)) 
			$$ = queueinst(pgm, @1, OPCODE_LDRg, $1, $2, 0, $3); 
		else 
			YYERROR; 
	}
	| T_STR REG MODIFIER '\n'		
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT)) 
			$$ = queueinst(pgm, @1, OPCODE_STRg, 0, $2, 0, $3); 
		else 
			YYERROR; 
	}
	| T_IOR REG '\n'		{ $$ = queueinst(pgm, @1, OPCODE_IORg, 0, $2, 0, 0);  }
	| T_IOR REG MODIFIER 	
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT)) 
			$$ = queueinst(pgm, @1, OPCODE_IORg, 0, $2, 0, $3); 
		else 
			YYERROR; 
	}
	| T_ISZ REG MODIFIER
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT)) 
			$$ = queueinst(pgm, @1, OPCODE_ISZg, 0, $2, 0, $3); 
		else 
			YYERROR; 
	}
	| T_AND REG '\n'		{ $$ = queueinst(pgm, @1, OPCODE_ANDg, 0, $2, 0, 0);  }
	| T_AND REG MODIFIER 	
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT)) 
			$$ = queueinst(pgm, @1, OPCODE_ANDg, 0, $2, 0, $3); 
		else 
			YYERROR; 
	}
	| T_DSZ REG MODIFIER		
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT)) 
			$$ = queueinst(pgm, @1, OPCODE_DSZg, 0, $2, 0, $3); 
		else 
			YYERROR; 
	}
	| T_JSM REG MODIFIER		
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT)) 
			$$ = queueinst(pgm, @1, OPCODE_JSMg, 0, $2, 0, $3); 
		else 
			YYERROR; 
	}
	| T_JMP REG '\n'		{ $$ = queueinst(pgm, @1, OPCODE_JMPg, 0, $2, 0, 0);  }
	| T_JMP REG MODIFIER 	
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT)) 
			$$ = queueinst(pgm, @1, OPCODE_JMPg, 0, $2, 0, $3); 
		else 
			YYERROR; 
	}
	| T_EXR '\n'		{ $$ = queueinst(pgm, @1, OPCODE_EXR, $1, 0, 0, 0);  }
	| T_DMA '\n'		{ $$ = queueinst(pgm, @1, OPCODE_DMA, 0, 0, 0, 0);   }
	| T_CMR '\n'		{ $$ = queueinst(pgm, @1, OPCODE_CMR, $1, 0, 0, 0);  }
	| T_TCR '\n'		{ $$ = queueinst(pgm, @1, OPCODE_TCR, $1, 0, 0, 0);  }

	/* Memory reference instruction group; modifier can be indirect */	

	| T_ADR expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_ADR, $1, 0, $2, 0); }
	| T_ADR expression MODIFIER '\n'	
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT))
			$$ = queueinst(pgm, @1, OPCODE_ADR, $1, 0, $2, $3);
		else
			YYERROR;
	}
	| T_CPR expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_CPR, $1, 0, $2, 0); }
	| T_CPR expression MODIFIER '\n'	
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT))
			$$ = queueinst(pgm, @1, OPCODE_CPR, $1, 0, $2, $3);
		else
			YYERROR;
	}
	| T_LDR expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_LDR, $1, 0, $2, 0); }
	| T_LDR expression MODIFIER '\n'	
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT))
			$$ = queueinst(pgm, @1, OPCODE_LDR, $1, 0, $2, $3);
		else
			YYERROR;
	}
	| T_STR expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_STR, $1, 0, $2, 0); }
	| T_STR expression MODIFIER '\n'	
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT))
			$$ = queueinst(pgm, @1, OPCODE_STR, $1, 0, $2, $3);
		else
			YYERROR;
	}
	| T_IOR expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_IOR, 0, 0, $2, 0);  }
	| T_IOR expression MODIFIER '\n'	
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT))
			$$ = queueinst(pgm, @1, OPCODE_IOR, 0, 0, $2, $3); 
		else
			YYERROR;
	}
	| T_ISZ expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_ISZ, 0, 0, $2, 0);  }
	| T_ISZ expression MODIFIER '\n'	
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT))
			$$ = queueinst(pgm, @1, OPCODE_ISZ, 0, 0, $2, $3); 
		else
			YYERROR;
	}
	| T_AND expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_AND, 0, 0, $2, 0);  }
	| T_AND expression MODIFIER '\n'	
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT))
			$$ = queueinst(pgm, @1, OPCODE_AND, 0, 0, $2, $3); 
		else
			YYERROR;
	}
	| T_DSZ expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_DSZ, 0, 0, $2, 0);  }
	| T_DSZ expression MODIFIER '\n'	
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT))
			$$ = queueinst(pgm, @1, OPCODE_DSZ, 0, 0, $2, $3); 
		else
			YYERROR;
	}
	| T_JSM expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_JSM, 0, 0, $2, 0);  }
	| T_JSM expression MODIFIER '\n'	
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT))
			$$ = queueinst(pgm, @1, OPCODE_JSM, 0, 0, $2, $3); 
		else
			YYERROR;
	}
	| T_JMP expression '\n'		{ $$ = queueinst(pgm, @1, OPCODE_JMP, 0, 0, $2, 0);  }
	| T_JMP expression MODIFIER '\n'	
	{ 
		if (checkmod(@1, $3, MOD_INDIRECT))
			$$ = queueinst(pgm, @1, OPCODE_JMP, 0, 0, $2, $3); 
		else
			YYERROR;
	}

	/* Pseudo-opcodes; assembler directives. */

	| T_ORG expression '\n'
	{ 
		int err = 0;
		int t;
		YYLTYPE eloc;

		t = evalexpr($2, 0, &err, &eloc);
		if (err) {
			yyerror(&eloc, NULL, "can't evaluate ORG expression: %s", errexpr(err));
			YYERROR;
		}

		if ((pgm->dot != -1) && (t < pgm->dot)) {
			yyerror(&@1, NULL, "ORG expression less than current address: current = 0%o, ORG value = 0%o",
				pgm->dot, t);
			YYERROR;
		} 

		$$ = queueinst(pgm, @1, OPCODE_ORG, 0, 0, 0, t);
	}
	| T_DW expression-list '\n'
	{ 
		Expr *e;
		Inst *i = NULL;

		for (e = $2; e != NULL; e = e->next) {
			i = queueinst(pgm, @1, OPCODE_DATA, 0, 0, e, 0);  
		}
		$$ = i;
	}
	| T_DB expression-list '\n'
	{
		Expr *e = $2;
		Expr *t, *result;
		Inst *i = NULL;

		/* DB is syntactic sugar for ((e & 0xff) << 8) | ((e->next) & 0xff) */ 

		while (e) {
			t = opexpr(@2, OP_SHL, opexpr(@2, OP_AND, e, numexpr(@2, 0xff)), numexpr(@2, 8));
			if (e->next) {
				result = opexpr(@2, OP_OR, t, opexpr(@2, OP_AND, e->next, numexpr(@2, 0xff)));
				e = e->next->next;
			} else {
				result = t;
				e = NULL;
			}
			i = queueinst(pgm, @1, OPCODE_DATA, 0, 0, result, 0);
		}
		$$ = i;
	}
	| T_ASC STRING '\n'
	{ 
		Expr *e;
		Inst *i = NULL;
		int n;

		/* this relies on a NUL byte at the end of the string */

		for (n = 0; n < strlen($2); n += 2) {
			e = numexpr(@2, $2[n]<<8 | $2[n+1]);
			i = queueinst(pgm, @1, OPCODE_DATA, 0, 0, e, 0);
		}

		$$ = i;
	}
	| T_BSS NUM '\n'
	{ 
		int n;
		Inst *i = NULL;
		Expr *e;

		if (pgm->dot == -1) {
			pgm->dot = DEFAULT_ORIGIN;
		}

		/* fill the space with zeros */

		e = numexpr(@2, 0);
		for (n = 0; n < $2; n++) {
			i = queueinst(pgm, @1, OPCODE_DATA, 0, 0, e, 0);
		}

		$$ = i;
	}
	| T_NOP '\n'
	{ 
		Inst *i = NULL;
		Expr *e;

		e = numexpr(@1, 0);
		i = queueinst(pgm, @1, OPCODE_DATA, 0, 0, e, 0);  
		$$ = i;
	}
	| '\n'
	{
		$$ = NULL;
	}
	;

expression-list:
	expression
	{
		$$ = $1;
	}
	| expression ',' expression-list
	{
		/* FIXME abstraction of Expr */
		$1->next = $3;
		$$ = $1;
	}
	;

expression:
	NUM { $$ = numexpr(@1, $1); }
	| LABEL 
	{ 
		Symbol *s;

		if ((s = findsymbol(pgm->st, $1, SYMBOL_FOLLOW)) == NULL) {
			s = addsymbol(pgm->st, $1, NULL, &@1);
		}

		$$ = symbolexpr(@1, s); 
	}
	| '.'	{ $$ = numexpr(@1, pgm->dot); }
	| expression '+' expression { $$ = opexpr(@2, OP_PLUS, $1, $3); }
	| expression '-' expression { $$ = opexpr(@2, OP_MINUS, $1, $3); }
	| expression '*' expression { $$ = opexpr(@2, OP_MUL, $1, $3); }
	| expression '/' expression { $$ = opexpr(@2, OP_DIV, $1, $3); }
	| expression '&' expression { $$ = opexpr(@2, OP_AND, $1, $3); }
	| expression '|' expression { $$ = opexpr(@2, OP_OR, $1, $3); }
	| expression '^' expression { $$ = opexpr(@2, OP_XOR, $1, $3); }
	| expression SHL expression { $$ = opexpr(@2, OP_SHL, $1, $3); }
	| expression SHR expression { $$ = opexpr(@2, OP_SHR, $1, $3); }
	| '-' expression  %prec NEG { $$ = opexpr(@1, OP_NEG, NULL, $2); } 
	| '~' expression  %prec NEG { $$ = opexpr(@1, OP_TWDL, NULL, $2); }
	| '(' expression ')'{ $$ = $2; }
	;
%%
