/* asm9830: a portable cross-assembler for the HP 9830A/B calculator */
/* Copyright (c) 2013-2014, Robert Ferguson (rob@bitscience.ca) */
/* All rights reserved; see LICENSE for details */

#include <stdio.h>
#include <stdlib.h> 
#include <string.h> 

#include "parser.tab.h"
#include "asm9830.h"

#define NBUCKET	23

STable *
createst(STable *parent)
{
	STable *st;

	if ((st = (STable *)malloc(sizeof(STable))) == NULL) 
		return(NULL);

	if ((st->table = createht(NBUCKET)) == NULL) {
		free(st);
		return(NULL);
	}

	st->parent = parent;

	return(st);
}

void
destroyst(STable *st)
{
	destroyht(st->table, 0);
	free(st);
}

/* symbol table listing requires a conveniently sortable table and a comparison function */

typedef struct _stlist {
	char *key;
	Symbol *s;
} STList;

static int
stlcmp(const void *l1, const void *l2)
{
	return(strcmp(((STList *)l1)->key, ((STList *)l2)->key));
}

void
listst(STable *st, FILE *fplist)
{
	Symbol *s;
	char *key;
	void *context = NULL;
	int count;
	STList *list;
	int i;

	count = countht(st->table);

	if ((list = (STList *)malloc(sizeof(STList) * count)) == NULL) {
		error("can't allocate STList");
		return;
	}

	i = 0;
	while((s = enumht(st->table, &key, &context))) {
		list[i].key = key;
		list[i].s = s;
		i++;
	}

	if (i != count) {
		error("corrupt symbol table; countht() and enumht() don't match");
		return;
	}

	qsort(list, count, sizeof(STList), stlcmp);

	for (i = 0; i < count; i++) {
		if (list[i].s->type == SYMBOL_FORWARD) 
			fprintf(fplist, "%32s:\tundefined\n", list[i].key);
		else
			fprintf(fplist, "%32s:\t%06o\n", list[i].key, list[i].s->value);
	}

	free(list);
}

Symbol *
addsymbol(STable *st, char *name, Symbol *existing, YYLTYPE *locp)
{
	Symbol *s;

	if (existing == NULL) {	
		if ((s = (Symbol *)malloc(sizeof(Symbol))) == NULL)
			return(NULL);

		s->type = SYMBOL_FORWARD;
		s->loc = *locp;
	} else {
		s = existing;
	}

	if (insertht(st->table, name, s) != s) {
		free(s);
		return(NULL);
	}
	return(s);
}

Symbol *
delsymbol(STable *st, char *name)
{
	return(deleteht(st->table, name));
}

Symbol *
findsymbol(STable *st, char *name, int follow)
{
	STable *p = st;
	Symbol *s;

	while (p) {
		if ((s = findht(p->table, name)))
			return(s);
		if (follow == SYMBOL_FOLLOW)
			p = p->parent;
		else
			return(NULL);
	}
	return(NULL);
}

Symbol *
setsymbol(STable *st, char *name, int value)
{
	Symbol *s;

	s = findht(st->table, name);

	if (s == NULL)
		return(NULL);

	s->value = value;
	s->type = SYMBOL_DEFINED;

	return(s);
}
